//*****************************************************************************
//
// qs-rgb.c - Quickstart for the EK-LM4F120XL Stellaris LaunchPad
//
// Copyright (c) 2012 Texas Instruments Incorporated.  All rights reserved.
// Software License Agreement
// 
// Texas Instruments (TI) is supplying this software for use solely and
// exclusively on TI's microcontroller products. The software is owned by
// TI and/or its suppliers, and is protected under applicable copyright
// laws. You may not combine this software with "viral" open-source
// software in order to form a larger program.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITH ALL FAULTS.
// NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT
// NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. TI SHALL NOT, UNDER ANY
// CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES, FOR ANY REASON WHATSOEVER.
// 
// This is part of revision 9453 of the EK-LM4F120XL Firmware Package.
//
//*****************************************************************************

#include <stdlib.h>
#include <math.h>
#include "inc/hw_types.h"
#include "inc/hw_memmap.h"
#include "inc/hw_hibernate.h"
#include "driverlib/fpu.h"
#include "driverlib/sysctl.h"
#include "driverlib/rom.h"
#include "driverlib/pin_map.h"
#include "driverlib/gpio.h"
#include "driverlib/systick.h"
#include "driverlib/interrupt.h"
#include "driverlib/hibernate.h"
#include "utils/uartstdio.h"
#include "drivers/rgb.h"
#include "drivers/buttons.h"
#include "main.h"
#include "ws2812.h"


//*****************************************************************************
//
// Array of pre-defined colors for use when buttons cause manual color steps.
//
//*****************************************************************************
static float fManualColors[7] = {0.0f, .214f, .428f, .642f, .856f, 1.07f, 
		1.284f};

//*****************************************************************************
//
// Application state structure.  Gets stored to hibernate memory for 
// preservation across hibernate events.  
//
//*****************************************************************************
volatile sAppState_t g_sAppState;

//*****************************************************************************
//
// The error routine that is called if the driver library encounters an error.
//
//*****************************************************************************
#ifdef DEBUG
void
__error__(char *pcFilename, unsigned long ulLine)
{
	UARTprintf("Assert failed in '%s', line : %d\n", pcFilename, (int)ulLine);
}
#endif


volatile unsigned long ticker = 0;


//*****************************************************************************
//
// Handler to manage the button press events and state machine transitions
// that result from those button events.
//
// This function is called by the SysTickIntHandler if a button event is 
// detected. Function will determine which button was pressed and tweak various
// elements of the global state structure accordingly.
//
//*****************************************************************************
void 
AppButtonHandler(void)
{
	static unsigned long ulTickCounter;

	ulTickCounter++;

	//
	// Switch statement to adjust the color wheel position based on buttons
	//
	switch(g_sAppState.ulButtons & ALL_BUTTONS)
	{

	case LEFT_BUTTON:
		//
		// Check if the button has been held long enough to peform another
		// color wheel increment.
		//
		if((ulTickCounter % APP_BUTTON_POLL_DIVIDER) == 0)
		{
			ws2812ledHasOVR[((LEDS_PHYS-1) * masterBrightness) / 100] = 0;
			if (masterBrightness > 10) masterBrightness -= 10; else masterBrightness = 1;
			ws2812ovrlayCounter = 200;		// 2 secs
			ws2812ledHasOVR[((LEDS_PHYS-1) * masterBrightness) / 100] = 1;

			//
			// Perform the increment and index wrap around.
			//
			g_sAppState.ulManualIndex++;
			if(g_sAppState.ulManualIndex >= APP_NUM_MANUAL_COLORS)
			{
				g_sAppState.ulManualIndex = 0;
			}
			g_sAppState.fColorWheelPos = APP_PI *
					fManualColors[g_sAppState.ulManualIndex];
		}

		//
		// Reset some state counts and system mode so that we know the user
		// is present and actively engaging with the application.
		//
		g_sAppState.ulModeTimer = 0;
		g_sAppState.ulMode = APP_MODE_NORMAL;
		break;

	case RIGHT_BUTTON:
		//
		// Check if the button has been held long enough to perform another
		// color wheel decrement.
		//
		if((ulTickCounter % APP_BUTTON_POLL_DIVIDER) == 0)
		{
			ws2812ledHasOVR[((LEDS_PHYS-1) * masterBrightness) / 100] = 0;
			if (masterBrightness < 90) masterBrightness += 10; else masterBrightness = 100;
			ws2812ovrlayCounter = 200;		// 2 secs
			ws2812ledHasOVR[((LEDS_PHYS-1) * masterBrightness) / 100] = 1;

			//
			// Perform the decrement and index wrap around.
			//
			if(g_sAppState.ulManualIndex == 0)
			{
				//
				// set to one greater than the last color so that we decrement
				// back into range with next instruction.
				//
				g_sAppState.ulManualIndex = APP_NUM_MANUAL_COLORS;
			}
			g_sAppState.ulManualIndex--;
			g_sAppState.fColorWheelPos = APP_PI *
					fManualColors[g_sAppState.ulManualIndex];
		}
		//
		// Reset some state counts and system mode so that we know the user
		// is present and actively engaging with the application.
		//
		g_sAppState.ulModeTimer = 0;
		g_sAppState.ulMode = APP_MODE_NORMAL;
		break;

	case ALL_BUTTONS:

		break;

	default:
	{
		//
		// Normal or remote mode and no user action will cause transition
		// to automatic scrolling mode.
		//
		if(g_sAppState.ulModeTimer < APP_AUTO_MODE_TIMEOUT)
		{
			g_sAppState.ulModeTimer++;
		}
		else
		{
			g_sAppState.ulMode = APP_MODE_AUTO;
		}

		//
		// reset the tick counter when no buttons are pressed
		// this makes the first button reaction speed quicker
		//
		ulTickCounter = APP_BUTTON_POLL_DIVIDER - 1;
	}
	break;
	}
}

//*****************************************************************************
//
// Uses the fColorWheelPos variable to update the color mix shown on the RGB
//
// ulForceUpdate when set forces a color update even if a color change
// has not been detected.  Used primarily at startup to init the color after
// a hibernate.
//
// This function is called by the SysTickIntHandler to update the colors on
// the RGB LED whenever a button or timeout event has changed the color wheel
// position.  Color is determined by a series of sine functions and conditions
//
//*****************************************************************************
void 
AppRainbow(unsigned long ulForceUpdate)
{
	static float fPrevPos;
	float fCurPos;
	float fTemp;
	volatile unsigned long * pulColors;

	pulColors = g_sAppState.ulColors;
	fCurPos = g_sAppState.fColorWheelPos;


	if((fCurPos != fPrevPos) || ulForceUpdate)
	{
		//
		// Preserve the new color wheel position
		//
		fPrevPos = fCurPos;

		//
		// Adjust the BLUE value based on the control state
		//
		fTemp = 65535.0f * sinf(fCurPos);
		if(fTemp < 0)
		{
			pulColors[GREEN] = 0;
		}
		else
		{
			pulColors[GREEN] = (unsigned long) fTemp;
		}


		//
		// Adjust the RED value based on the control state
		//
		fTemp = 65535.0f * sinf(fCurPos - APP_PI / 2.0f);
		if(fTemp < 0)
		{
			pulColors[BLUE] = 0;
		}
		else
		{
			pulColors[BLUE] = (unsigned long) fTemp;
		}


		//
		// Adjust the GREEN value based on the control state
		//
		if(fCurPos < APP_PI)
		{
			fTemp = 65535.0f * sinf(fCurPos + APP_PI * 0.5f);
		}
		else
		{
			fTemp = 65535.0f * sinf(fCurPos + APP_PI);
		}
		if(fTemp < 0)
		{
			pulColors[RED] = 0;
		}
		else
		{
			pulColors[RED] = (unsigned long) fTemp;
		}

		//
		// Update the actual LED state
		//
		RGBColorSet(pulColors);
	}

}

//*****************************************************************************
//
// Called by the NVIC as a result of SysTick Timer rollover interrupt flag
//
// Checks buttons and calls AppButtonHandler to manage button events.
// Tracks time and auto mode color stepping.  Calls AppRainbow to implement
// RGB color changes.
//
//*****************************************************************************
void 
SysTickIntHandler(void)
{

	static float x;

	ticker++;
	if (ws2812ovrlayCounter > 0)
		ws2812ovrlayCounter--;

	g_sAppState.ulButtons = ButtonsPoll(0,0);
	AppButtonHandler();

	//
	// Auto increment the color wheel if in the AUTO mode. AUTO mode is when
	// device is active but user interaction has timed out.
	//
	if(g_sAppState.ulMode == APP_MODE_AUTO)
	{
		g_sAppState.fColorWheelPos += APP_AUTO_COLOR_STEP;
	}

	//
	// Provide wrap around of the control variable from 0 to 1.5 times PI
	//
	if(g_sAppState.fColorWheelPos > (APP_PI * 1.5f))
	{
		g_sAppState.fColorWheelPos = 0.0f;
	}
	if(x < 0.0f)
	{
		g_sAppState.fColorWheelPos = APP_PI * 1.5f;
	}

	//
	//    Set the RGB Color based on current control variable value.
	//
	AppRainbow(0);


}



void BouncingBall(void)
{
	unsigned long t;
	int i;
	int ballPos = 0;
	int ballDir = 1;
	rgbValue_t ballColor;
#define TAIL_LEN	6
	uint16_t	tailMult[TAIL_LEN] = {100, 60, 30, 15, 5, 0};	// percent values of tail

	srand(3591);

#define COL_CHG		((lrand48()-RAND_MAX/2) < 0 ? -32 : 32)

	for (i = 0; i < LEDS_PHYS; i++)
	{
		ws2812ledRGB[i].R = 0;
		ws2812ledRGB[i].G = 0;
		ws2812ledRGB[i].B = 0;
	}

	ballColor.R = 0x80;
	ballColor.G = 0x80;
	ballColor.B = 0x80;

	t = 0;

	while (1)
	{
		if (t < ticker)
		{
			t = ticker + (50 / LEDS_PHYS) + 1;
			i = (int)ballColor.R + COL_CHG;
			ballColor.R = putInRange(i, 32, 255);
			i = (int)ballColor.G + COL_CHG;
			ballColor.G = putInRange(i, 32, 255);
			i = (int)ballColor.B + COL_CHG;
			ballColor.B = putInRange(i, 32, 255);

			i = RAND_MAX;

			if (ballDir > 0)
			{
				if (++ballPos >= LEDS_PHYS)
				{
					ballPos = LEDS_PHYS-1;
					ballDir = 0;
				}
				else
				{
					for (i = 0; i < TAIL_LEN; i++)
					{
						int j = ballPos - i;
						if (j >= 0)
						{
							ws2812ledRGB[j].R = ((int)ballColor.R * tailMult[i]) / 100;
							ws2812ledRGB[j].G = ((int)ballColor.G * tailMult[i]) / 100;
							ws2812ledRGB[j].B = ((int)ballColor.B * tailMult[i]) / 100;
						}
					}
				}
			}
			else
			{
				if (--ballPos <= 0)
				{
					ballPos = 0;
					ballDir = 1;
				}
				else
				{
					for (i = 0; i < TAIL_LEN; i++)
					{
						int j = ballPos + i;
						if (j < LEDS_PHYS)
						{
							ws2812ledRGB[j].R = ((int)ballColor.R * tailMult[i]) / 100;
							ws2812ledRGB[j].G = ((int)ballColor.G * tailMult[i]) / 100;
							ws2812ledRGB[j].B = ((int)ballColor.B * tailMult[i]) / 100;
						}
					}
				}
			}

			while (ledBusy)
				;

			WS2812update();
		}
	}
}




//*****************************************************************************
//
// Main function performs init and manages system.
//
// Called automatically after the system and compiler pre-init sequences.
// Performs system init calls, restores state from hibernate if needed and 
// then manages the application context duties of the system.
//
//*****************************************************************************
int main(void)
{
	unsigned long ulResetCause;

	//
	// Enable stacking for interrupt handlers.  This allows floating-point
	// instructions to be used within interrupt handlers, but at the expense of
	// extra stack usage.
	//
	FPUEnable();
	FPUStackingEnable();

	//
	// Set the system clock to run at 80Mhz off PLL with external crystal as
	// reference.
	//
	//    SysCtlClockSet(SYSCTL_SYSDIV_5 | SYSCTL_USE_PLL | SYSCTL_XTAL_16MHZ | SYSCTL_OSC_MAIN);
	SysCtlClockSet(SYSCTL_SYSDIV_2_5 | SYSCTL_USE_PLL | SYSCTL_OSC_MAIN | SYSCTL_XTAL_16MHZ);  //80MHz

	//
	// Enable the hibernate module
	//
	SysCtlPeripheralEnable(SYSCTL_PERIPH_HIBERNATE);

	//
	// Enable and Initialize the UART.
	//
	SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOA);
	GPIOPinConfigure(GPIO_PA0_U0RX);
	GPIOPinConfigure(GPIO_PA1_U0TX);
	GPIOPinTypeUART(GPIO_PORTA_BASE, GPIO_PIN_0 | GPIO_PIN_1);
	UARTStdioInit(0);

	UARTprintf("Welcome to the Stellaris LM4F120 LaunchPad!\n");
	UARTprintf("Type 'help' for a list of commands\n");
	UARTprintf("> ");

	//
	// Determine why system reset occurred and respond accordingly.
	//
	ulResetCause = SysCtlResetCauseGet();
	SysCtlResetCauseClear(ulResetCause);
	{
		//
		// Reset was do to a cold first time power up.
		//
		UARTprintf("Power on reset. Hibernate not active.\n");
		UARTprintf("> ");

		g_sAppState.ulMode = APP_MODE_NORMAL;
		g_sAppState.fColorWheelPos = 0;
		g_sAppState.fIntensity = APP_INTENSITY_DEFAULT;
		g_sAppState.ulButtons = 0;
	}


	//
	// Initialize the RGB LED. AppRainbow typically only called from interrupt
	// context. Safe to call here to force initial color update because
	// interrupts are not yet enabled.
	//
	RGBInit(0);
	RGBIntensitySet(g_sAppState.fIntensity);
	AppRainbow(1);
	RGBEnable();

	//
	// Initialize the buttons
	//
	ButtonsInit();

	//
	// Initialize the SysTick interrupt to process colors and buttons.
	//
	SysTickPeriodSet(SysCtlClockGet() / APP_SYSTICKS_PER_SEC);
	SysTickEnable();
	SysTickIntEnable();
	IntMasterEnable();

	WS2812init();

	BouncingBall();

	return 0;
}
