/******************************************************
 *
 *	Driver for WS2812 led strips; Uses SPI port
 *
 *	(c) Peter Schulten (pitschu), Mlheim, Germany
 *	peter_(at)_pitschu.de
 *
 *	Die unvernderte Wiedergabe und Verteilung dieses gesamten Sourcecodes
 *	in beliebiger Form ist gestattet, sofern obiger Hinweis erhalten bleibt.
 *
 * 	Ich stelle diesen Sourcecode kostenlos zur Verfgung und biete daher weder
 *	Support an noch garantiere ich fr seine Funktionsfhigkeit. Auerdem
 *	bernehme ich keine Haftung fr die Folgen seiner Nutzung.

 *	Der Sourcecode darf nur zu privaten Zwecken verwendet und modifiziert werden.
 *	Darber hinaus gehende Verwendung bedarf meiner Zustimmung.
 *
 *	History
 *	Nov 2013	pitschu		Start of work
 *
*/
#include "stdio.h"
#include "stdint.h"
#include <math.h>
#include "inc/hw_types.h"
#include "inc/hw_memmap.h"
#include "inc/hw_ints.h"
#include "inc/hw_gpio.h"
#include "inc/hw_ssi.h"
#include "driverlib/udma.h"
#include "driverlib/sysctl.h"
#include "driverlib/pin_map.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/ssi.h"

#include "ws2812.h"



rgbValue_t 				ws2812ledRGB[LEDS_PHYS];			// color data for all leds
rgbValue_t 				ws2812ledOVR[LEDS_PHYS];			// overlay color data for all leds (has higher prio than ws2812ledRGB
volatile unsigned char 	ws2812ledHasOVR[LEDS_PHYS];
volatile unsigned long 	ws2812ovrlayCounter;				// ignore overlay when 0 (decr in system ticker)

static uint8_t 			ws2812dmaBuffer[WS2812_DMABUF_LEN];	// buffer for SSI/dma, one byte per bit + reset pulse
volatile uint8_t		ledBusy = 0;							// = 1 while dma is sending data to leds

volatile uint16_t		masterBrightness = 50;


unsigned char ucControlTable[1024] __attribute__ ((aligned(1024)));		// DMA control tab


//-------------------------------------------------------------------------------------------------------------

void SSI2IRQandler (void)
{
	unsigned long irqStat = SSIIntStatus(WS2812_SSI_BASE, false);

	SSIIntClear(WS2812_SSI_BASE, 0x0F);		// clear all flags

	if (irqStat)
	{

		if (uDMAIntStatus() & (1 << WS2812_DMA_CH))		// irq causd by DMA
		{
		HWREG(GPIO_PORTB_BASE + (GPIO_O_DATA + (GPIO_PIN_4 << 2))) = GPIO_PIN_4;		// set port for OSZI
			tDMAControlTable *controlTable = (tDMAControlTable *)uDMAControlBaseGet();

			controlTable = &controlTable[WS2812_DMA_CH];

			uDMAIntClear (1 << WS2812_DMA_CH);

			unsigned long transferSize = (unsigned long)(&ws2812dmaBuffer[WS2812_DMABUF_LEN-1]) - (unsigned long)(controlTable->pvSrcEndAddr);

			if (transferSize > 0)		// continue DMA
			{
				uDMAChannelTransferSet(WS2812_DMA_CH | UDMA_PRI_SELECT,		// select primary entry in control tab
						UDMA_MODE_BASIC,									// use basic transfer mode
						(void *)(controlTable->pvSrcEndAddr + 1),			// source is the next byte after last transfer
						(void *)(WS2812_SSI_BASE + SSI_O_DR),				// destination is SSI2 data register
						min(WS2812_DMA_TRANSSIZE, transferSize));			// number of transfers depends on LED strip length

				uDMAChannelEnable(WS2812_DMA_CH);		// start transmission
			}
			else		// DMA finished, all bytes transfered
			{
				ledBusy = 0;							// get ready for next transfer
			}

		}
		HWREG(GPIO_PORTB_BASE + (GPIO_O_DATA + (GPIO_PIN_4 << 2))) = 0;				// clear debug port
	}

}


// convert a RGB byte into SPI data stream with 2 bits per byte
static uint8_t *rgb2pwm(uint8_t *dest, const uint8_t color)
{
	register int count;
	register uint32_t data = color;

	for (count = 0; count < 4; count++)
	{
		switch (data & 0xC0)
		{
		case  0x00:	*dest = WS_BITS_00; break;
		case  0x40:	*dest = WS_BITS_01; break;
		case  0x80:	*dest = WS_BITS_10; break;
		case  0xC0:	*dest = WS_BITS_11; break;
		}

		dest++;
		data <<= 2;
	}

	return dest;
}




// convert color codes from "ws2812_framebuffer" into serial "commands" for SPI
int WS2812update(void)
{
	register uint32_t i;
	register rgbValue_t *r;
	uint8_t * bufp = ws2812dmaBuffer;
	register uint32_t c;

	for (i = 0; i < LEDS_PHYS; i++)
	{
		if (ws2812ovrlayCounter && ws2812ledHasOVR[i])
		{
			r = (rgbValue_t *)&ws2812ledOVR[i];
			bufp = rgb2pwm(bufp, r->G);
			bufp = rgb2pwm(bufp, r->R);
			bufp = rgb2pwm(bufp, r->B);
		}
		else
		{
			r = (rgbValue_t *)&ws2812ledRGB[i];
			c = ((uint32_t)r->G * masterBrightness) / 100;
			bufp = rgb2pwm(bufp, (const uint8_t)c);
			c = ((uint32_t)r->R * masterBrightness) / 100;
			bufp = rgb2pwm(bufp, (const uint8_t)c);
			c = ((uint32_t)r->B * masterBrightness) / 100;
			bufp = rgb2pwm(bufp, (const uint8_t)c);

			ws2812ledHasOVR[i] = 0;
		}
	}

	return WS2812startDMA();		// send it to RGB stripe
}




void WS2812init(void)
{
	int i;

	for (i = 0; i < (WS2812_DMABUF_LEN - WS2812_RESET_LEN); i++)
		ws2812dmaBuffer[i] = WS_BITS_00;
	for (; i < WS2812_DMABUF_LEN; i++)
		ws2812dmaBuffer[i] = 0x00;		// should hold the line LOW for at least 50usec

	for (i = 0; i < LEDS_PHYS; i++)
	{
		ws2812ledRGB[i].B = 0;
		ws2812ledRGB[i].G = 0;
		ws2812ledRGB[i].R = 0;

		// set standard overlay GREEN for all LEDs but clear 'has ovrlay flag'
		ws2812ledOVR[i].B = 0;
		ws2812ledOVR[i].G = 0xFF;
		ws2812ledOVR[i].R = 0;
		ws2812ledHasOVR[i] = 0;
	}


	SysCtlPeripheralEnable(WS2812_SSI_PORTPERIPH);		// enable the IO parts; we use PB7 as SPI-2 out
	SysCtlPeripheralEnable(WS2812_SSI_PERIPH);
	SysCtlPeripheralEnable(SYSCTL_PERIPH_UDMA);

	// use PB4 for debugging
	GPIODirModeSet(GPIO_PORTB_BASE, GPIO_PIN_4, GPIO_DIR_MODE_OUT);
	GPIOPadConfigSet(GPIO_PORTB_BASE, GPIO_PIN_4, GPIO_STRENGTH_8MA_SC, GPIO_PIN_TYPE_STD);
	GPIOPinWrite(GPIO_PORTB_BASE, GPIO_PIN_4, 0x00);	// set port for testing
	GPIOPinWrite(GPIO_PORTB_BASE, GPIO_PIN_4, GPIO_PIN_4);	// set port for testing
	SysCtlDelay(25L);
	GPIOPinWrite(GPIO_PORTB_BASE, GPIO_PIN_4, 0x00);	// set port for testing

	//----------------------------------

	GPIOPinTypeSSI(GPIO_PORTB_BASE, WS2812_SSI_PORTPIN);
	GPIOPadConfigSet(GPIO_PORTB_BASE, WS2812_SSI_PORTPIN, GPIO_STRENGTH_8MA_SC, GPIO_PIN_TYPE_OD);
	GPIOPinConfigure(WS2812_GPIO_SSI);			// connect pin to SSI

	SSIDisable (WS2812_SSI_BASE);
	SSIClockSourceSet(WS2812_SSI_BASE, SSI_CLOCK_SYSTEM);

	// use Freescale SPI Frame Format with SPO=0 and SPH=1
	SSIConfigSetExpClk (WS2812_SSI_BASE,
			SysCtlClockGet(),
			SSI_FRF_MOTO_MODE_1,				// should be the freescale mode 0/1
			SSI_MODE_MASTER,
			3200000UL,							// 3,2 Mhz
			8);									// 8 bits for 2 WS2812 bits

	SSIIntDisable(WS2812_SSI_BASE, 0xFF);				// no IRQ from SSI
	IntEnable(WS2812_INT_SSI);						// but DMA uses the SSI irq line
	IntPrioritySet(WS2812_INT_SSI, 1);

	uDMAEnable();

	uDMAControlBaseSet((void*)&ucControlTable[0]);		// set base addr of chan control tab (is in RAM s.o.)

	uDMAChannelAssign (WS2812_DMA_ASSIGN);	// assign the DMA chan to the SSI2 interface

	uDMAChannelAttributeDisable(WS2812_DMA_CH,
			UDMA_ATTR_USEBURST |
			UDMA_ATTR_ALTSELECT |
			UDMA_ATTR_REQMASK);	// clear all unused attributes
	uDMAChannelAttributeEnable (WS2812_DMA_CH, UDMA_ATTR_HIGH_PRIORITY);

	uDMAChannelControlSet(WS2812_DMA_CH | UDMA_PRI_SELECT,	//	select primary entry in control tab
			UDMA_SIZE_8 |
			UDMA_SRC_INC_8 |
			UDMA_DST_INC_NONE |	// prepare for 8 bit
			UDMA_ARB_4);		// FIFO size of SSI

	SSIDMAEnable (WS2812_SSI_BASE, SSI_DMA_TX);
	SSIEnable(WS2812_SSI_BASE);

	ledBusy = 0;

	WS2812startDMA();
}




// transfer framebuffer data to SSI-TX port
int WS2812startDMA(void)
{
	if (ledBusy)		// last DMA is not finished
		return 1;

	ledBusy = 1;

	uDMAChannelTransferSet(WS2812_DMA_CH | UDMA_PRI_SELECT,		// select primary entry in control tab
			UDMA_MODE_BASIC,									// use basic transfer mode
			(void *) &ws2812dmaBuffer[0],						// source is the prepared array (look at rgb2pwm())
			(void *)(WS2812_SSI_BASE + SSI_O_DR),				// destination is SSI2 data register
			min(WS2812_DMA_TRANSSIZE, WS2812_DMABUF_LEN));		// number of transfers depends on LED strip length

	uDMAChannelEnable(WS2812_DMA_CH);		// starts transmission

	return 0;
}

